from rql.utils import register_function
from logilab.database import FunctionDescr

from cubicweb import Binary

options = (
    ('s3-auto-delete', {
        'type': 'yn',
        'default': True,
        'help': 'Delete S3 objects on entity deletion',
        'group': 's3',
        'level': 2}),
    )


class STKEY(FunctionDescr):
    """return the S3 key of the bytes attribute stored using the S3 Storage (s3s)
    """
    rtype = 'Bytes'

    def update_cb_stack(self, stack):
        assert len(stack) == 1
        stack[0] = self.source_execute

    def as_sql(self, backend, args):
        raise NotImplementedError(
            'This callback is only available for S3Storage '
            'managed attribute. Is STKEY() argument S3S managed?')

    def source_execute(self, source, session, value):
        s3key = source.binary_to_str(value)
        try:
            return Binary(s3key)
        except OSError as ex:
            source.critical("can't read %s: %s", s3key, ex)
            return None


register_function(STKEY)



import re
from contextlib import contextmanager

from six import PY3

from cubicweb.server.sources import storages
from cubicweb.devtools.testlib import CubicWebTC
from cubicweb import Binary
from cubicweb.server.migractions import ServerMigrationHelper
from cubicweb_s3storage import testing


def create_image(cnx, data=b'the-data', **kwargs):
    return cnx.create_entity('Image', data=Binary(data), **kwargs)


class S3StorageTC(testing.S3StorageTestMixin, CubicWebTC):

    def test_s3key_gen(self):
        s3storage = self.repo.system_source.storage('Image', 'data')
        with self.admin_access.client_cnx() as cnx:
            fobj = create_image(cnx, b'some content')
            cnx.commit()
            eid = fobj.eid
            k1 = s3storage.get_s3_key(fobj, 'data')
        with self.admin_access.client_cnx() as cnx:
            fobj = cnx.find('Image', eid=eid).one()
            k2 = s3storage.get_s3_key(fobj, 'data')
        self.assertEqual(k1, k2)

    def test_entity_create(self):
        with self.admin_access.client_cnx() as cnx:
            eid = create_image(cnx, b'some content').eid
            cnx.commit()

            key = cnx.execute('Any STKEY(D) WHERE X is Image, X data D, '
                              'X eid %(eid)s', {'eid': eid}).rows[0][0]
            key = key.getvalue().decode()

        data = self.s3_bucket.Object(key).get()['Body'].read()
        self.assertEqual(data, b'some content')

    def test_entity_modify(self):
        with self.admin_access.client_cnx() as cnx:
            eid = create_image(cnx, b'some content').eid
            cnx.commit()

            key = cnx.execute('Any STKEY(D) WHERE X is Image, X data D, '
                              'X eid %(eid)s', {'eid': eid}).rows[0][0]
            key = key.getvalue().decode()
        with self.admin_access.client_cnx() as cnx:
            fobj = cnx.find('Image', eid=eid).one()
            fobj.cw_set(data=Binary(b'something else'))
            cnx.commit()

        data = self.s3_bucket.Object(key).get()['Body'].read()
        self.assertEqual(data, b'something else')

    def test_entity_retrieve(self):
        binstuff = ''.join(chr(x) for x in range(256))
        if PY3:
            binstuff = binstuff.encode()
        with self.admin_access.client_cnx() as cnx:
            eid = create_image(cnx, binstuff).eid
            cnx.commit()

        with self.admin_access.client_cnx() as cnx:
            rset = cnx.execute('Any D WHERE F eid %(eid)s, F data D',
                               {'eid': eid})
            self.assertTrue(rset)
            data = rset.rows[0][0]
        self.assertEqual(data.read(), binstuff)

    def test_entity_delete(self):
        self.repo.vreg.config['s3-auto-delete'] = True
        with self.admin_access.client_cnx() as cnx:
            eid = create_image(cnx, b'some content').eid
            cnx.commit()

            key = cnx.execute('Any STKEY(D) WHERE X is Image, X data D, '
                              'X eid %(eid)s', {'eid': eid}).rows[0][0]
            key = key.getvalue().decode()
        keys = [x.key for x in self.s3_bucket.objects.all()]
        self.assertIn(key, keys)
        with self.admin_access.client_cnx() as cnx:
            cnx.execute('DELETE Image X WHERE X eid %(eid)s', {'eid': eid})
            cnx.commit()
        keys = [x.key for x in self.s3_bucket.objects.all()]
        self.assertNotIn(key, keys)


class S3StorageMigrationTC(testing.S3StorageTestMixin, CubicWebTC):

    @contextmanager
    def mh(self):
        with self.admin_access.repo_cnx() as cnx:
            yield cnx, ServerMigrationHelper(
                self.repo.config, self.repo.schema,
                repo=self.repo, cnx=cnx,
                interactive=False)

    def test_entity_migration(self):
        with self.admin_access.client_cnx() as cnx:
            create_image(cnx, thumbnail=Binary(b'some content'))
            cnx.commit()

        # Re-use storage instance of "data" attribute as it already has s3
        # mock activated.
        s3_storage = self.repo.system_source.storage('Image', 'data')
        with self.mh() as (cnx, mh):
            storages.set_attribute_storage(
                self.repo, 'Image', 'thumbnail', s3_storage)

            mh.cmd_storage_changed('Image', 'thumbnail')
            cnx.commit()

        with self.admin_access.client_cnx() as cnx:
            key = cnx.execute('Any STKEY(D) WHERE X is Image, '
                              'X thumbnail D').rows[0][0]
            key = key.getvalue().decode()
            # check it looks like an UUID generated by uuid.uuid1()
            # sorry, I'm lazy, this regex is a bit too permissive...
            self.assertTrue(re.match(r'\w{8}-\w{4}-\w{4}-\w{4}-\w{12}', key))

            value = cnx.execute('Any D WHERE X is Image, '
                                'X thumbnail D').rows[0][0]
            self.assertEqual(b'some content', value.getvalue())


if __name__ == '__main__':
    from unittest import main
    main()




from cubicweb.server.hook import Hook
from cubicweb_s3storage.storages import S3Storage


class S3StorageStartupHook(Hook):
  __regid__ = 's3tests.server-startup-hook'
  events = ('server_startup', 'server_maintenance')

  def __call__(self):
      storage = S3Storage('test-bucket')
      self.repo.system_source.set_storage('Image', 'data', storage)



# pylint: disable=W0622
"""cubicweb-s3storage application packaging information"""


modname = 'cubicweb_s3storage'
distname = 'cubicweb-s3storage'

numversion = (1, 0, 2)
version = '.'.join(str(num) for num in numversion)

license = 'LGPL'
author = 'LOGILAB S.A. (Paris, FRANCE)'
author_email = 'contact@logilab.fr'
description = 'A Cubicweb Storage that stores the data on S3'
web = 'http://www.cubicweb.org/project/%s' % distname

__depends__ = {
    'cubicweb': '>= 3.24.7',
    'six': '>= 1.4.0',
    'boto3': None,
}
__recommends__ = {}

classifiers = [
    'Environment :: Web Environment',
    'Framework :: CubicWeb',
    'Programming Language :: Python',
    'Programming Language :: JavaScript',
]



import boto3
from mock import patch
from moto import mock_s3


class S3StorageTestMixin(object):

    s3_bucket = 'test-bucket'

    def setUp(self):
        s3_mock = mock_s3()
        s3_mock.start()
        resource = boto3.resource('s3', region_name='us-east-1')
        self.s3_bucket = resource.create_bucket(Bucket=self.s3_bucket)
        patched_storage_s3_client = patch(
            'cubicweb_s3storage.storages.S3Storage._s3_client',
            return_value=boto3.client('s3'),
        )
        patched_storage_s3_client.start()
        self._mocks = [
            s3_mock,
            patched_storage_s3_client,
        ]
        super(S3StorageTestMixin, self).setUp()

    def tearDown(self):
        super(S3StorageTestMixin, self).tearDown()
        while self._mocks:
            self._mocks.pop().stop()
